import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.Iterator;
import java.util.List;
import java.awt.image.BufferedImage;
import java.awt.Color;

/**
 * Jumble an image within a Fifteen puzzle and
 * allow the user to try to solve it.
 * The scenario stops when the puzzle is solved.
 * 
 * @author David J. Barnes
 * @version 2008.08.26
 */
public class FifteenWorld extends World
{
    // The number of tiles per side.
    public static final int SQUARE_SIDE = 4;
    
    // The width of the tiles. This could be set from the image size.
    private static final int TILE_WIDTH = 50;
    // The image to appear on the tiles.
    private static final String IMAGE_FILENAME = "elvis.jpg";
    // How many times to swap the hole with adjacent tiles at the
    // start of the game.
    private static final int RANDOMIZER = 1000;
    
    // The hole.
    private Hole hole;

    /**
     * Constructor for objects of class FifteenWorld.
     * 
     */
    public FifteenWorld()
    {    
        // Create a new world.
        super(SQUARE_SIDE, SQUARE_SIDE, TILE_WIDTH);
        // Create the tiles and hole.
        populate();
        // Jumble up the puzzle.
        jumble();
    }
    
    /**
     * Create the tiles and hole.
     * The tiles are decorated with a chopped-up image.
     */
    public void populate()
    {
        hole = new Hole();
        hole.setImage(createHoleImage(TILE_WIDTH));
        
        // The image to be distributed over the tiles.
        GreenfootImage image = new GreenfootImage(IMAGE_FILENAME);
        image.scale(SQUARE_SIDE * TILE_WIDTH, SQUARE_SIDE * TILE_WIDTH);
        BufferedImage mainImage = image.getAwtImage();
        // Fill the grid with tiles.
        for(int row = 0; row < SQUARE_SIDE; row++) {
            for(int col = 0; col < SQUARE_SIDE; col++) {
                Tile tile = new Tile(hole);
                tile.setImage(createTileImage(row, col, mainImage));
                addObject(tile, col, row);
            }
        }
        
        // Remove the tile from the bottom right-hand corner and
        // replace it with a Hole.
        int holeX = SQUARE_SIDE - 1, holeY = SQUARE_SIDE - 1;
        removeObjects(getObjectsAt(holeX, holeY, Tile.class));
        addObject(hole, holeX, holeY);
    }
    
    /**
     * Check for the end of the game when all tiles are in position.
     */
    public void act()
    {
        List<Actor> tiles = getObjects(Tile.class);
        boolean allInPosition = true;
        Iterator<Actor> it = tiles.iterator();
        while(allInPosition && it.hasNext()) {
            Tile tile = (Tile) it.next();
            if(!tile.inPosition()) {
                allInPosition = false;
            }
        }
        if(allInPosition) {
            Greenfoot.playSound("cheer2.aif");
            Greenfoot.stopSimulation();
        }        
    }
    
    /**
     * Perform a series of moves of the hole to randomize the
     * initial position.
     */
    public void jumble()
    {
        for(int i = 0; i < RANDOMIZER; i++) {
            List<Tile> tiles = hole.getAdjacentTiles();
            int index = Greenfoot.getRandomNumber(tiles.size());
            Tile t = tiles.get(index);
            if(t.nextToHole()) {
                t.switchWithHole();
            }
        }
    }
    
    /**
     * Create an image for the tile at the given position, from
     * the provided image.
     * @param row The row of the tile.
     * @param col The col of the tile.
     * @param mainImage The image to be segmented.
     * @return The image for the tile.
     */
    private GreenfootImage createTileImage(int row, int col, BufferedImage mainImage)
    {
        GreenfootImage image = new GreenfootImage(TILE_WIDTH, TILE_WIDTH);
        int rowOffset = row * TILE_WIDTH;
        int colOffset = col * TILE_WIDTH;
        for(int x = 0; x < TILE_WIDTH; x++) {
            for(int y = 0; y < TILE_WIDTH; y++) {
                int pixel = mainImage.getRGB(colOffset + x, rowOffset + y);
                image.setColorAt(x, y, new Color(pixel));
            }
        }
        // Draw a border around it.
        image.setColor(Color.BLACK);
        image.drawRect(0, 0, TILE_WIDTH - 1, TILE_WIDTH - 1);
        return image;
    }
    
    /**
     * Create an image for the hole.
     * @param width The width of the image.
     * @return The image for the hole.
     */
    private GreenfootImage createHoleImage(int width)
    {
        GreenfootImage image = new GreenfootImage(width, width);
        image.setColor(Color.WHITE);
        image.fillRect(0, 0, width, width);
        image.setColor(Color.BLACK);
        image.drawRect(0, 0, width - 1, width - 1);
        return image;
    }
}
